// work in progress - not yet working
// attempt to port from BASIC to c
// copyright belongs to mikroelektronica for the original code and to Yves Bourdon who wrote BASIC PIC version

// Still to do:
//   add in more options for display of other message types and detailed dignostic info on demand
//   work out why LCD works OK for 8MHz, but not 32MHz clock  (external LCD OK, but on-board display fails to initialise)
//   much RAM wasted on text messages - should change to using ROM, but I just can't be bothered :(
//   country code shows as 8, but should be 12 for UK

// European Program Type decode (Pty) used, but reality does not seem to fit spec - beyond me what to make of it all

//' * Project name:
//'     FM_Click
//' * Revision History:
//'     20130520:
//'       - initial release (DO from Mikroelektronica);
//'     20130920
//'     This project is a simple demonstration of working with FM tunner Si4703
//'     module FM Click
//' * Test configuration:
//'     MCU:             PIC18F45K22
//'     dev.board:       Easy PICv7
//'     Oscillator:      HS 8 MHz, 8.0000 MHz Crystal
//'     ext. modules:    fm click board :
//' * Developpement tool : MikroC Pro for PIC v7.2.0
//' * Based on MikroElektronica FM_Click example for ARM
//' * Display 4x20 LCD on Port B or alternatively standard ME 16x2 display on-board
//' * Harware Setup:
//'     - J5 MUST BE SET ON 3.3V !!!!!!
//'     - Place FM click into mikroBUS socket 1 on EasyPIC v7
//'     - SW3 : 0N/0N/OFF/ON/OFF/OFF/OFF/OFF
//'     - SW4 : BCK ON
//'     - PORTA : RA1 et RA2 pull up
//      - PORTD : pull up on all pins
//      - JP7 to ground position
//'     - All other switches : default
//'
//' * Switch used
//' RD0 and RD1 : Seek up and down
//' RD2 and RD3 : Volume up and down
//' RD4 and RD5 : select what to display on LCD line 2
//'
//' * Display line 1 : RDS sation name, Frequency (Europe) + STEREO/MONO
//' * Display line 2 : Rssi + Volume output (default) - other options selectable with buttons on PORTD 4 and 5
//' * Display line 3 : Clock and date decoded from RDS data
//' * Display line 4 : Rssi + Volume output
//'
//' Note:
//' * You may adjust block_err_sum if RDS display is too long to come up.
// */

//  IEC 62106-4 paper gives details on RDS.
//  tested in South England, but PTY information for North America used, rather than Europe, since it seems to fit reality better

#define FM_SEN LATE0_bit
#define FM_RST LATE1_bit
#define FM_SDIO LATC4_bit
#define FM_SCLK LATC3_bit
#define FM_GP1 RB0_bit
#define FM_GP2 LATA2_bit

// Lcd pinout settings
sbit LCD_RS at RB4_bit;
sbit LCD_EN at RB5_bit;
sbit LCD_D7 at RB3_bit;
sbit LCD_D6 at RB2_bit;
sbit LCD_D5 at RB1_bit;
sbit LCD_D4 at RB0_bit;

// Pin direction
sbit LCD_RS_Direction at TRISB4_bit;
sbit LCD_EN_Direction at TRISB5_bit;
sbit LCD_D7_Direction at TRISB3_bit;
sbit LCD_D6_Direction at TRISB2_bit;
sbit LCD_D5_Direction at TRISB1_bit;
sbit LCD_D4_Direction at TRISB0_bit;

#define Si4703_address 0x20                                                     // Si4703 address for I2C is 0x10 (bits 1 to 7) with bit 0 = 0 for write, 1 for read (i.e 0x20=write, 0x21=read)

#define FM_all_Reg 0x0F                                                         // cause read of all 16 registers of 16 bits each
#define FM_0a0b_Reg 2                                                           // cause read of only 2 registers

signed char powerOn;                                                            // powerOn = 0
unsigned char Rssi;                                                             // signal strength
unsigned int freq;                                                              // frequency FM
unsigned char volume;                                                           // current volume
unsigned char RDS_ok;                                                           // RDS data valid flag
unsigned char country_code;                                                     // RDS country code
unsigned char coverage;                                                         // RDS coverage
unsigned char program_ref;                                                      // RDS program reference
unsigned int ChipID;                                                            // chip identity number (0x242 for Si4703B19)
unsigned char SelectD;                                                          // line 2 display selection
unsigned char old_SelectD;                                                      // prevent selecting more than once
unsigned char text5[7];                                                         // general purpose variable
unsigned int index;                                                             // general purpose variable
unsigned char rds_pty;                                                          // 5 bits program type
unsigned int block_err_name;                                                    // error count on station name blocks
unsigned char block_err_sum;                                                    // sum of all block errors
unsigned char stereo;                                                           // =1 stereo, 0= mono
unsigned char str_size;                                                         // number of characters in RDS name string
unsigned char flag_RDS;                                                         // flag set to display RDS info
unsigned char flag_Clock;                                                       // flag set to display time and date information
unsigned char RDS_text[9];                                                      // station name
unsigned char debug_text[11];                                                   // general purpose text buffer for conversions
unsigned char info_text[65];                                                    // RDS text 32 or 64 characters
unsigned char local_hour;                                                       // time display calculation registers
unsigned char local_minute;
unsigned char local_day;
unsigned char local_month;
unsigned char local_year;
unsigned char len_month[13] = {0, 30, 28, 31, 30, 31, 30, 31, 31, 30, 31, 31, 31};         // number of days in each month

// for some strange reason, FM chip read returns reg 10 first - order as below
// write starts from reg 2  - confusing, but true - see datasheet :(
unsigned char shadowReg[32];                                                    // Copy of 16 registers (16 bits each register) of Si4703
// 0000 Reg 10                                                                  shadowreg  0,1
// 0000 Reg 11                                                                  shadowReg  2,3
// 0000 Reg 12 RDS A                                                            shadowreg  4,5
// 0000 Reg 13 RDS B                                                            shadowReg  6,7
// 0000 Reg 14 RDS C                                                            shadowreg  8,9
// 0000 Reg 15 RDS D                                                            shadowreg 10,11
// 1242 Reg  0 Device ID                                                        shadowreg 12,13
// 1200 Reg  1 Chip ID                                                          shadowreg 14,15
// 0000 Reg  2 Power Config                                                     shadowreg 16,17
// 0000 Reg  3                                                                  shadowreg 18,19
// 0000 Reg  4                                                                  shadowreg 20,21
// 0000 Reg  5                                                                  shadowreg 22,23
// 0000 Reg  6                                                                  shadowreg 24,25
// 0100 Reg  7 Test1 set bit 15 for crystal oscillator                          shadowreg 26,27
// 0000 Reg  8                                                                  shadowreg 28,29
// 0000 Reg  9                                                                  shadowreg 30,31

// update shadow registers by either setting or clearing bits
void modShadowReg(unsigned char reg, unsigned int value, unsigned char set_reset){

unsigned char reg_pos;                                                          // Index of register byte in shadow register
                                                                                // Reading of Si4703 registers start from 0x0A-0x0F and then starts from 0x00-0x09
    reg_pos = 12 + (reg << 1);                                                  // Calculate position of register which needs to be modified  (Reg.0x00 is in position 12)
    if (set_reset){                                                             // Sets the desired bits in register
        shadowReg[reg_pos] |= value>>8;                                         // Split value to higher
        reg_pos++;
        shadowReg[reg_pos] |=  (value&0xff);                                    // and lower byte
    }
    else{                                                                       // Resets the desired bits in register
        shadowReg[reg_pos] &=  ( ~ ( value>>8 ) );                              // Split value to higher
        reg_pos++;
        shadowReg[reg_pos] &=  ( ~ ( value &0xff) );                            // and lower byte
    }
}

// send shodow register(s) to Si4703 chip
void FM_Write(unsigned char reg, unsigned int value, unsigned char set_reset){
unsigned char wr_cnt;
unsigned char ii;
  
    wr_cnt = ( reg - 1 ) << 1;                                                  // Calculate how many bytes need to be written
    modShadowReg(reg, value, set_reset);                                        // Modify register bytes
    I2C1_Start();
    I2C1_wr(Si4703_address);
    for (ii = 16; ii <= (16 + wr_cnt);ii++){                                    // ex : pour registre 0x07, on crit 12 octets (addresse + data)  partir du registre 0x02 soit @shadowReg(16)
        I2C1_Wr(shadowReg[ii]);                                                 // Writing to register starts from 0x02 register
    }
    I2C1_stop();
}

// read either all 16 Si4703 registers (16 bits each) or just registers 10 and 11
void FM_Read(unsigned char cmd){                                                // Reading of registers starts from 0x0A register
unsigned char ii;

    if(cmd == FM_all_Reg){                                                      // 0x0f to read all 16 registers of 16 bits each
        I2C1_Start();
        I2C1_wr(Si4703_address + 1 );                                           // write address 0x10 plus write bit gives 0x21
        for(ii = 0; ii <= 30; ii++){
            shadowReg[ii] = I2C1_rd(1);                                         // read next byte, send ack
        }
        shadowReg[31] = I2C1_rd(0);                                             // send  nack  on last data byte only
        I2C1_Stop();
    }

    if(cmd == FM_0a0b_Reg){                                                     // Read only STATUS and READCHAN registers  (except it seems to get regs 0 and 1 - not a and b
        I2C1_Start();
        I2C1_wr(Si4703_address +1 );
        for( ii = 0; ii <= 2; ii++){
          shadowReg[ii] = I2C1_rd(1);
        }
        shadowReg[3] = I2C1_rd(0);
        I2C1_Stop();
    }
}

// change volume
void SetVolume(unsigned char vol){
    if ( ( vol >= 0 ) && ( vol <= 16) ) {
        vol = vol & 0x0F;
        modShadowReg(0x05, 0x000F, 0);
        FM_Write(0x05, vol, 1);                                                 // Write Volume value to SYSCONFIG2 register
        delay_ms(5);
    }
}

// get frequency and Rssi information
void  Refresh_Freq_Rds_Rssi(){
unsigned char block_errA, block_errB, block_errC, block_errD;
  
    FM_Read(FM_0a0b_Reg);
    freq = shadowReg[3] + 875;                                                  // Begining of Europe band 87.5 Mhz = 875 * 100kHz
    rssi = shadowReg[1];
    stereo = shadowReg[0] & 0b00000001;                                         // reg10 bit 0 set for stereo
    RDS_ok = (shadowReg[0] & 0b10000000) >> 7;                                  // reg 10 bit 15 is RDS ready bit
    block_errA = shadowReg[0];                                                  // reg 10 bits 8 to 15 (block_errA not used in this program)
    block_errB = (shadowReg[2] & 0b11000000 ) >> 6;                             // reg 11 bits 15, 14
    block_errC = (shadowReg[2] & 0b00110000 ) >> 4;                             // reg 11 bits 13, 12
    block_errD = (shadowReg[2] & 0b00001100 ) >> 2;                             // reg 11 bits 11, 10
    block_err_name = block_errD;                                                // errrors in radio station name
    block_err_sum = block_errB + block_errC + block_errD;                       // sum up all block errors into one variable
    country_code = shadowReg[0] >> 4;                                           // block 0 bits 15 to 12 contain country code
    coverage = shadowReg[0] & 0x0f;                                             // block 0 bits 11 to 8 contain coverage number
    program_ref = shadowReg[1];                                                 // b;ock 0 bits 7 to 0 contain program reference number
}

// display frequency and associated information
void display_fm(){
    if(stereo){                                                                 // test stereo flag
        lcd_out(1,16,"ST");
    }                                                                           // Affiche indicateur st/mono
    else{
        lcd_out(1,16,"  ");
    }
    wordtostr(freq,text5);                                                      // covert binary to ASCII
    lcd_chr(1,14,text5[4]);
    lcd_chr(1,13,'.');
    lcd_chr(1,12,text5[3]);
    lcd_chr(1,11,text5[2]);
    lcd_chr(1,10,text5[1]);
    lcd_out(4,1,"Rssi: ");                                                      // Display  Rssi (signal strength
    Lcd_Chr(4,7,(rssi / 10) + 48);
    Lcd_Chr(4,8,(rssi % 10) + 48);
    lcd_out(4,14,"Vol:");                                                       // Display Volume
    Lcd_Chr(4,19,(volume / 10) + 48);
    Lcd_Chr(4,20,(volume % 10) + 48);

    if(SelectD == 0){                                                           // selection 0 is for Rssi and vol info on LCD line 2
         Lcd_Chr(2,7,(rssi / 10) + 48);
         Lcd_Chr(2,8,(rssi % 10) + 48);
         Lcd_Chr(2,15,(volume / 10) + 48);
         Lcd_Chr(2,16,(volume % 10) + 48);
    }
    if(SelectD == 2){                                                           // selection 2 is for program type info on LCD line 2
         lcd_out(2,1,"Pty ");
         Lcd_Chr(2,5,(rds_pty / 10) + 48);
         Lcd_Chr(2,6,(rds_pty % 10) + 48);
         switch(rds_pty){                                                       // European program type numbering - not all decodes given to save memory
             case 1:
                 Lcd_Out(2, 8, "news     ");
                 break;
             case 5:
                 Lcd_Out(2, 8, "education");
                 break;
             case 6:
                 Lcd_Out(2, 8, "drama    ");
                 break;
             case 7:
                 Lcd_Out(2, 8, "culture  ");
                 break;
             case 8:
                 Lcd_Out(2, 8, "science  ");
                 break;
             case 9:
                 Lcd_Out(2, 8, "varied   ");
                 break;
             case 10:
                 Lcd_Out(2, 8, "pop music");
                 break;
             case 11:
                 Lcd_Out(2, 8, "rock     ");
                 break;
             case 14:
                 Lcd_Out(2, 8, "classical");
                 break;
             default:
                 Lcd_Out(2, 8, "         ");
                 break;
         }
    }
    if(SelectD == 4){                                                           // selection 4 is for group type info on LCD line 2
         ByteToStr(country_code,debug_text);
         Lcd_Out(2, 14, debug_text);
    }
    if(SelectD == 5){                                                           // selection 5 is for coverage info on LCD line 2
         ByteToStr(coverage,debug_text);
         Lcd_Out(2, 14, debug_text);
    }
    if(SelectD == 6){                                                           // selection 6 is for program reference num on LCD line 2
         ByteToStr(program_ref,debug_text);
         Lcd_Out(2, 14, debug_text);
    }
}


void FM_Tune(unsigned int freq){
unsigned int calc;
signed char STC_bit;
    STC_bit = 0;
    calc = freq;
    calc = calc - 875;
    calc = calc | 0x8000;                                                       // Set TUNE Flag
    if (powerON == 1){                                                          // Check for power state of FM click
        FM_Read(FM_0a0b_Reg);
        modShadowReg(0x03, 0x83FF, 0);
        FM_Write(0x03, calc, 1);                                                // Write modifiled frequency value  + activate tune
        Delay_ms(60);
        while (STC_bit == 0){
            FM_Read(FM_0a0b_Reg);
            STC_bit = shadowReg[0] & 0x40;
        }
        FM_Read(FM_0a0b_Reg);
        modShadowReg(0x03, 0x8000, 0);
        FM_Write(0x03, 0x8000, 0);                                              // Set TUNE = 0
        while (STC_bit > 0){
            FM_Read(FM_all_Reg);
            STC_bit = shadowReg[0] & 0x40;
        }
    }
    else{
        ModShadowReg(0x03, calc, 1);                                            // If power is off, only modify the shadow register
    }
}


void FM_Seek(unsigned char direction){
signed char STC_bit, SF_BL_bit;
    if (direction == 1){
        modShadowReg(0x02, 0x0300, 0);
        FM_Write(0x02, 0x0300, 1);                                              // Activate SEEK direction up and Seek ON
    }
    else{
        modShadowReg(0x02, 0x0300, 0);
        FM_Write(0x02, 0x0100, 1);                                              // Activate SEEK direction down and Seek ON
    }
    STC_bit = 0 ;
    while( STC_bit == 0 ){
        FM_Read(FM_0a0b_Reg);
        STC_bit = (shadowReg[0] & 0b01000000) >> 6;
        delay_ms(60);
    }
    modShadowReg(0x02, 0x0300, 0);
    FM_Write(0x02, 0x0100, 0);                                                  // Deactivate SEEK,keep
    while( STC_bit == 1 ){
        FM_Read(FM_0a0b_Reg);
        STC_bit = (shadowReg[0] & 0b01000000) >> 6;
    }
    FM_Read(FM_0a0b_Reg);
    SF_BL_bit = (shadowReg[0] & 0b00100000) >> 5;                               // 0 pas de station 1 station ok
}

void Rds_scan(){
unsigned char Ah, Al, Bh, Bl, Ch, Cl, Dh, Dl;                                   // RDS information arrives in registers 12 to 15 RDSA to RDSD
unsigned char pos;
unsigned char service;                                                          // shows what kind of message is contained within RDS registers
signed int utc_offset;                                                          // adjust for local time offset to UTC
signed int rds_hours;                                                           // 5 bits hours
signed int rds_minutes;                                                         // 6 bits minutes
unsigned long mjulian_days;                                                     // modified julian days
unsigned long wdrds;                                                            // Jour de la semaine calcul dans la routine RDS
unsigned char leap;

    RDS_text[8] = 0;                                                            // ensure null terminator
    
    if(RDS_ok){

        FM_Read(FM_all_Reg);
        Ah = shadowReg[4];                                                      // reg 12 contains RDS block A
        Al = shadowReg[5];
        Bh = shadowReg[6];                                                      // reg 13 contains RDS block B
        Bl = shadowReg[7];
        Ch = shadowReg[8];                                                      // reg 14 contains RDS block C
        Cl = shadowReg[9];
        Dh = shadowReg[10];                                                     // reg 15 contains RDS block D
        Dl = shadowReg[11];
        service = Bh & 0b11111000;                                              // mask to get GTYPE and find out what kind of data is in RDS message
        
        if (service == 0){                                                      // group type 0A or 0B station name
            if (block_err_name < 3){                                            // number of errors acceptable in station name - adjust for quality verses speed
                pos = Bl & 3;                                                   // read C0 C1 to find which two station name characters of 8
                if( pos == 0){                                                  // new message - since first 2 characters
                   if( str_size == 8){                                          // On peut imprimer le message car on a reu 8 caractres
                       flag_RDS = 1;
                   }
                   else{                                                        // On n'a pas encore 8 caractres
                       if( RDS_text[0] == 0){                                   // On ne rempli qu'une fois !
                           RDS_text[0] = Dh;                                    // On commence  remplir les 2 premiers caractres
                           RDS_text[1] = Dl;
                           str_size = str_size + 2;
                       }
                   }
                }
                else{
                    if( (pos == 1) && (RDS_text[2] == 0)){                      // second pair of characters
                        RDS_text[2] = Dh;
                        RDS_text[3] = Dl;
                        str_size = str_size + 2;
                    }
                    if( (pos == 2) && (RDS_text[4] == 0)){
                        RDS_text[4] = Dh;                                       // third pair of characters
                        RDS_text[5] = Dl;
                        str_size = str_size + 2;
                    }
                    if( (pos == 3) && (RDS_text[6] == 0)){
                        RDS_text[6] = Dh;                                       // fourth pair of characters
                        RDS_text[7] = Dl;
                        str_size = str_size + 2;
                    }
                }
                rds_pty = (((Bh & 3) *4)+(Bl>>5) & 0x1f);                       // 5 bits program type
            }
        }

        // group type 2A and 2B are for radio text (32 or 64 characters of programme information)
        if((service == 0x20) || (service == 28)){                               // group type 2A, 2B radio text
            if( block_err_sum < 1){                                             // number of errors acceptable for all text blocks
                pos = BL & 0x0f;                                                // determine which group of 4 characters
                info_text[pos*4] = CH;
                info_text[pos*4 + 1] = CL;
                info_text[pos*4 + 2] = DH;
                info_text[pos*4 + 3] = DL;
                if(selectD == 8){
                    for( pos = 0 ; pos < 16;pos++){
                        LCD_Chr(2,pos+1,info_text[pos]);
                    }
                }
                if(selectD == 9){
                    for( pos = 0 ; pos < 16;pos++){
                        LCD_Chr(2,pos+1,info_text[pos+16]);
                    }
                }
                if(selectD == 10){
                    for( pos = 0 ; pos < 16;pos++){
                        LCD_Chr(2,pos+1,info_text[pos+32]);
                    }
                }
                if(selectD == 11){
                    for( pos = 0 ; pos < 16;pos++){
                        LCD_Chr(2,pos+1,info_text[pos+48]);
                    }
                }
            }
        }

        // handle time and date - date not currently working
        if( service == 0x40){                                                   // Group Type 4A, time and date (%01001010)
            if( block_err_sum == 0){                                            // number of errors acceptable for all time/date blocks
                mjulian_days = (32768ul * (Bl & 0x03)) + (128 * Ch) + (Cl >> 1); // days since midnight on November 17, 1858
                utc_offset = (Dl & 0x1f) >> 1;                                  // read half-hours and change to hours
                rds_minutes = ((Dh & 0x0f) << 2) + ((Dl & 0xc0) >> 6);          // 6 bits minutes
                rds_hours = ((Cl & 0x01) << 4) + ((Dh & 0xf0) >> 4);            // 5 bits hours
                if (Dl & 0x20){                                                 // test bit 5 which is 0 for add local offset to UTC, 1 for subtract
                    rds_hours -= utc_offset;
                }
                else{
                    rds_hours += utc_offset;
                }
                if( rds_hours >= 24){                                           // handle overflow
                    rds_hours -= 24;
                    mjulian_days += 1;                                          // adjust date if passing midnight
                }
                if( rds_hours < 0){                                             // handle underflow
                    rds_hours += 24;
                    mjulian_days += 1;                                          // adjust date if passing midnight
                }
                
                // handle date - convert modified julian days to readable format
                // this is a really dumb way to calculate date, but simple is often good
                if(SelectD == 3){                                               // only calculate modified julian days if display selected
                    longwordtostr(mjulian_days,debug_text);
                    lcd_out(2,2,debug_text);                                    // display modified Julian Days (since November 17 1858)
                }
                if(SelectD == 1){                                               // only calculate date if display selected
                    mjulian_days -= 58119ul;                                    // change to days since 1/1/2018 (MJD 58119) to make calculation simpler
                    local_year = 18;
                    local_month = 1;
                    local_day = 1;
                    leap = 0;

                    // count through days and months until zero
                    while(mjulian_days > 31){                                   // bug waiting to suface here - surely !!!
                        if(mjulian_days > (len_month[local_month] + leap)){
                             mjulian_days -= (len_month[local_month] + leap);
                             local_month += 1;
                             if(local_month > 12){
                                 local_month = 1;
                                 local_year += 1;
                             }
                             leap = 0;                                              // start by assuming no leap day
                             if (local_month == 2){                                 // possible leap day in February
                                 if(local_year == 20) leap = 1;
                                 if(local_year == 24) leap = 1;
                                 if(local_year == 28) leap = 1;
                                 if(local_year == 32) leap = 1;
                                 if(local_year == 36) leap = 1;
                                 if(local_year == 40) leap = 1;
                                 if(local_year == 42) leap = 1;                     // I dont plan this program to be in use after 2045
                             }
                        }
                        local_day = mjulian_days;
                    }
                }
                
                
                local_minute = rds_minutes;
                local_hour = rds_hours;
                flag_Clock = 1;                                                 // Clock updated
            }
        }
    }
}

unsigned char FM_Activate(){
unsigned int active;

    active = ChipID;
    FM_Read(FM_all_Reg);
    modShadowReg(0x02, 0x4801, 0);
    FM_Write(0x02, 0x4801, 1);                                                  // Write ENABLE bit, disable Mute , RDS mode = Verbose
    Delay_ms(200);
    while (active == ChipID){                                                   // Wait for module powerup
        FM_Read(FM_all_Reg);
        active = shadowReg[14] << 8;
        active = active | shadowReg[15];
    }
    modShadowReg(0x04, 0xFFFF, 0);                                              // Write default Volume and RSSI Seek Threshold values
    FM_Write(0x04,0x1800,1);                                                    // Enable RDS + De-emphasis Europe
    modShadowReg(0x05, 0xFFFF, 0);                                              // Write default values
    FM_Write(0x05, 0x101F, 1);                                                  // RSSI seek Threshold = 0x10, band = Europe, Channel spacing = 100kHZ, Volume = max
    Delay_ms(110);                                                              // Max delay
    powerOn = 1;                                                                // Update powerOn flag
    return shadowReg[0];
}

void FM_Deactivate(){
    modShadowReg(0x02 , 0x4000, 0 );
    FM_Write(0x02, 0x0040, 1);
    powerOn = 0;                                                                // Updare powerOn flag
}

// bring FM chip out of reset (with control pins in correct state for 2-wire I2C communication)
void FMClick_Init(){
    FM_RST = 1;                                                                 // RE1 Init chip done - 2wires interface due to SDIO=0 et SDEN=1)
    Delay_ms(1);
    I2C1_Init(100000);                                                          // Initialize I2C module for communication with FM Click board
    FM_Read(FM_all_Reg);                                                        // read all 16 registers (16 bits each) to shadow_reg[]
    ChipID = shadowReg[14] * 256;                                               // REG 0 is the 14th byte to be returned by chip. Save Chip ID before powerup.
    ChipID = ChipID | shadowReg[15];                                            // expect chip ID of 0x1242
    Delay_ms(1);
    modShadowReg(0x07, 0x0081, 0);                                              // register 7, value 0081, set/reset = 'and'
    FM_Write(0x07, 0x8100, 1);                                                  // Activate Crystal Oscillator reg7 bit15 =1, bit8=1 to preserve power-on value , set/reset = 'or'
    Delay_ms(500);
}

void init_RDS (unsigned char efface){
    if (efface == 1){
        lcd_out(3,1,"                    ");                                      // clear time/date display
        if(SelectD == 2){
            lcd_out(2,1,"                    ");                                  // clear LCD line 2 if program type on display
        }
        lcd_out(1,1,"        ");                                                  // clear radio name
        rds_pty = 0;
    }
    RDS_text[0] = 0;
    for (index = 0; index <= 7 ; index++){
        RDS_text[index] = 0;
    }
    flag_RDS = 0;                                                               // No text to display
    flag_Clock = 0;                                                             // No time and date to display
    str_size = 0;                                                               // No charactere received
}

void display_time (unsigned char col ){
    if(SelectD == 1){                                                           // selection 1 displays time and date on line 2
        Lcd_Chr(2,col,local_hour / 10   + 48);                                  // display hour on LCD
        Lcd_Chr(2,col+1,(local_hour % 10)   + 48);
        Lcd_chr(2,col+2,':');
        Lcd_Chr(2,col+3,(local_minute / 10) + 48);
        Lcd_Chr(2,col+4,(local_minute % 10) + 48);
        Lcd_chr(2,col+5,' ');
    }
}

void display_date (unsigned char col ){
    if(SelectD == 1){                                                           // selection 1 displays time and date on line 2
        Lcd_Chr(2,col, (local_day / 10)   + 48);                                // display date
        Lcd_Chr(2,col+1, (local_day % 10)   + 48);
        Lcd_chr(2,col+2,'-');
        Lcd_Chr(2,col+3, (local_month / 10)   + 48);
        Lcd_Chr(2,col+4, (local_month % 10)   + 48);
        Lcd_out(2,col+5,"-20");
        Lcd_Chr(2,col+8, (local_year / 10)  + 48);
        Lcd_Chr(2,col+9, (local_year % 10)  + 48);
    }
}

// clear out info_text array
void Clear_Info(){
char pos;

    for(pos = 0; pos < 65 ; pos++){
        info_text[pos] = ' ';
    }
}

// crude way to de-bouce switches - just a simple delay - also affects repetition rate
void debounce(){
    Delay_ms(100);
}

// main program start
void main(){

     PLLEN_bit = 0;                                                             // turn off PLL, since LCD will not work at 32MHz
     
     ANSELA = 0;                                                                // Configure PORTB pins as digital
     ANSELB = 0;                                                                // Configure PORTB pins as digital
     ANSELC = 0;                                                                // Configure PORTB pins as digital
     ANSELD = 0;                                                                // Configure PORTB pins as digital
     ANSELE = 0;                                                                // Configure PORTB pins as digital
     SLRCON = 0;                                                                // All ports standard slew
     C1ON_bit = 0;                                                              // Disable comparators
     C2ON_bit = 0;
     TRISD = 0xff;                                                              // all inputs for buttons and switch selections

     TRISA0_bit = 0;                                                            // Output : led test
     LATA0_bit = 1;                                                             // Alimente LED
     TRISA1_bit = 1;                                                            // Bouton selection menu
     TRISA2_bit = 1;                                                            // GPIO2 input

     TRISE1_bit = 0;                                                            // #RST
     FM_RST = 0;                                                                // activate #RST to disable si4703 radio chip
     Delay_ms(1);
     TRISE0_bit = 0;                                                            // SEN
     FM_SEN = 1;                                                                // = 1
     Delay_ms(1);
     TRISC4_bit = 0;                                                            // SDIO
     FM_SDIO = 0;                                                               // 2 wires interface  (SDIO = 0 et SEN = 1 on rising edge of Reset)
     Delay_ms(1);
     TRISC3_bit = 0;                                                            // SCLK
     FM_SCLK = 0;                                                               // Pas d'importance tant que Reset = 0
     Delay_ms(1);

     Delay_ms(200);                                                             // allow plenty of time for LCD to power up

     Lcd_Init();                                                                // Initialize Lcd
     Lcd_Cmd(_LCD_CLEAR);                                                       // Clear display
     Lcd_Cmd(_LCD_CURSOR_OFF);                                                  // Cursor off

     Lcd_out(1,1,"BOURDON FM Click");
     FMClick_Init();
     Lcd_out(1,1,"init.. ");
     FM_activate();
     lcd_out(1,1,"    Wait Freq...");
     FM_Tune(1036);
     lcd_out(1,1,"      Tuned !   ");

     SelectD = 0;                                                               // default line 2 information display
     old_SelectD = 0xff;
     volume = 2;                                                                // Default volume
     setVolume(volume);
     delay_ms(5);

     delay_ms(2000);
     lcd_cmd(_lcd_clear);
     init_RDS(0);                                                               // Initialisation du RDS sans effacement des infos
     rds_pty = 0;

     while(1){

         if( !PORTD.B0){                                                        // low when button pressed
             init_RDS(1);                                                       // Chgt station donc init RDS et effacement des infos
             FM_Seek(1);                                                        // Seek up
             Debounce();                                                        // crude way to de-bounce
         }

         if( !PORTD.B1){
             init_RDS(1);                                                       // Chgt station donc init RDS et effacement des infos
             FM_Seek(0);                                                        // Seek down
             Debounce();                                                        // crude way to de-bounce
         }

         if( !PORTD.B2){                                                        // Volume up
             if( volume <15){
                 volume = volume + 1;
             }
             Setvolume(volume);
             Debounce();                                                        // crude way to de-bounce
         }

         if( !PORTD.B3){                                                        // volume down
             if( volume > 0){
                 volume = volume - 1;
             }
             SetVolume(volume);
             Debounce();                                                        // crude way to de-bounce
         }

         if( !PORTD.B4){                                                        // display selection up
            SelectD++;
             Debounce();                                                        // crude way to de-bounce
         }

         if( !PORTD.B5){                                                        // display selection down
             SelectD--;
             Debounce();                                                        // crude way to de-bounce
         }

         Refresh_Freq_Rds_Rssi();
         display_fm();                                                          // Display : Freq + Stereo/mono + Rssi + volume
         Rds_scan();
         if( flag_RDS == 1){                                                    // On a bien reu 8 caractres et donc on affiche
             lcd_out(1,1,RDS_text);
             init_RDS(0);
         }
         if(flag_Clock){
             display_time(1);
             display_date(7);
         }

         LATA0_bit = ~LATA0_bit;                                                // Toggle A0 to prove that program is still running

         if(old_SelectD != SelectD){                                            // prevent making the same selection twice
             SelectD &= 0x0f;                                                   // limit to 16 options
             old_SelectD = SelectD;                                             // allow only one update per selection
             switch(SelectD){                                                   // line 2 information display selection
                 case 0:{
                     LCD_Out(2,1,"Rssi:    Vol:   ");
                     break;
                 }
                 case 1:{
                     LCD_Out(2,1,"Time and Date...");
                     break;
                 }
                 case 2:{
                     LCD_Out(2,1,"Pty             ");
                     break;
                 }
                 case 3:{
                     LCD_Out(2,1," mod Julian days");
                     break;
                 }
                 case 4:{
                     LCD_Out(2,1,"country code    ");
                     break;
                 }
                 case 5:{
                     LCD_Out(2,1,"area coverage   ");
                     break;
                 }
                 case 6:{
                     LCD_Out(2,1,"Program ref     ");
                     break;
                 }
                 case 8:{
                     LCD_Out(2,1," Info text 0    ");
                     Clear_Info();
                     break;
                 }
                 case 9:{
                     LCD_Out(2,1," Info text 1    ");
                     Clear_Info();
                     break;
                 }
                 case 10:{
                     LCD_Out(2,1," Info text 2    ");
                     Clear_Info();
                     break;
                 }
                 case 11:{
                     LCD_Out(2,1," Info text 3    ");
                     Clear_Info();
                     break;
                 }
                 default:{
                     LCD_Out(2,1,"Version 290618b ");
                     break;
                 }
             }
         }
    }

}